#include "stdio.h"
#include "stdlib.h"
#include "XC3dWaveprop.h"
#include "XC3dAcouMod.h"

//////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////
int XC3dAcouMod_Init( int ntin,  int nxin,  int nyin,  int nzin, int nsin,  \
                      float otin,float oxin,float oyin,float ozin,          \
                      float dtin,float dxin,float dyin,float dzin,          \ 
                      int nxtapein, int nytapein, int nztapein,             \ 
                      float* sdepin, float *rdepin, float* velin,           \
                      int needwvmvin, int jtmvin, int freesurf )
{
    int ix,iy,iz,stat;
    float temp,v,vmax;
    // take input parameters
    XC3dAcouMod_nt=ntin;
    XC3dAcouMod_nx=nxin;
    XC3dAcouMod_ny=nyin;
    XC3dAcouMod_nz=nzin;
    XC3dAcouMod_ns=nsin;

    XC3dAcouMod_ot=otin;
    XC3dAcouMod_ox=oxin;
    XC3dAcouMod_oy=oyin;
    XC3dAcouMod_oz=ozin;

    XC3dAcouMod_dt=dtin;
    XC3dAcouMod_dx=dxin;
    XC3dAcouMod_dy=dyin;
    XC3dAcouMod_dz=dzin;

    XC3dAcouMod_nxtape=nxtapein;
    XC3dAcouMod_nytape=nytapein;
    XC3dAcouMod_nztape=nztapein;

    XC3dAcouMod_sdep=sdepin;
    XC3dAcouMod_rdep=rdepin;
    XC3dAcouMod_vel=velin;

    // determine if padded region is reasonable
    if (nxin<=2*nxtapein)
       WpErrmsg("Error, nx less or equal than twice of padded x reigon\n")
    if (nyin<=2*nytapein)
       WpErrmsg("Error, ny less or equal than twice of padded y reigon\n")
    if (nzin<=2*nztapein)
       WpErrmsg("Error, nz less or equal than twice of padded z reigon\n")

    // determine if dx==dy==dz
    if ( XC3dAcouMod_dx!=XC3dAcouMod_dy )
       Errmsg("ERROR, dx not equal to dy, exit!!!\n")
    else if ( XC3dAcouMod_dy!=XC3dAcouMod_dz )
       Errmsg("ERROR, dy not equal to dz, exit!!!\n")
    else if (  XC3dAcouMod_dz!=XC3dAcouMod_dx )
       Errmsg("ERROR, dz not equal to dx, exit!!!\n")
    else 
    {
       // produce (vel*dt/dx)**2 and get vmax/izmax_inc on the way
       vmax=0.;
       XC3dAcouMod_isdep=(int*)malloc(nxin*nyin*sizeof(int));
       XC3dAcouMod_irdep=(int*)malloc(nxin*nyin*sizeof(int));
       XC3dAcouMod_veldtinvdx2=(float*)malloc(nxin*nyin*nzin*sizeof(float));
       for (iy=0;iy<nyin;iy++)
          for (ix=0;ix<nxin;ix++)
             for (iz=0;iz<nzin;iz++)
             {
                v=XC3dAcouMod_vel[ zxyloc(iz,ix,iy) ];
                temp= v * XC3dAcouMod_dt/XC3dAcouMod_dx;
                XC3dAcouMod_veldtinvdx2[ zxyloc(iz,ix,iy) ]=temp*temp;
                if ( v  > vmax)
                   vmax = v;
                if (temp>0.86)
                  Errmsg2("ERROR, stability condition for wavepropagation is violated, temp=%f which is larger than 0.86\n",temp)
             }
       XC3dAcouMod_izmax_inc= vmax*XC3dAcouMod_dt/XC3dAcouMod_dz + 1;
    }
    // initilize wave propagation module
    stat=XC3dWaveprop_Init(ntin,nxin,nyin,nzin,nxtapein,nytapein,nztapein,needwvmvin, jtmvin, freesurf);
    if (stat==1)
       Errmsg("Error in XC3dWaveprop_Init, program terminated!\n")

    return 0;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////
int XC3dAcouMod_op(int tsign, int is, float *src, float *dat, float *wvmv)
{
  float *sdep,*rdep;
  int ix, iy;
  int nx, ny;
  int izmax_inc,stat;
  float dz;

  VerbMsg("Calculating sdep\n")
  sdep=&XC3dAcouMod_sdep[is*XC3dAcouMod_nx*XC3dAcouMod_ny];
  rdep=&XC3dAcouMod_rdep[is*XC3dAcouMod_nx*XC3dAcouMod_ny];
 
  nx=XC3dAcouMod_nx;
  ny=XC3dAcouMod_ny;
  dz=XC3dAcouMod_dz;
  // convert float depth (sdep, rdep) into integer index depth(isdep, irdep) 
 
  VerbMsg("Calculating isdep\n")
  for (iy=0;iy<ny;iy++)
     for (ix=0;ix<nx;ix++)
        XC3dAcouMod_isdep[ xyloc(ix,iy) ]=(sdep[ xyloc(ix,iy) ]-XC3dAcouMod_oz)/dz+XC3dAcouMod_nztape;

  VerbMsg("Calculating irdep\n")
  for (iy=0;iy<ny;iy++)
     for (ix=0;ix<nx;ix++)
        XC3dAcouMod_irdep[ xyloc(ix,iy) ]=(rdep[ xyloc(ix,iy) ]-XC3dAcouMod_oz)/dz+XC3dAcouMod_nztape;


  // wave propgation
  izmax_inc=XC3dAcouMod_izmax_inc;
  stat=XC3dWaveprop_op0( tsign, XC3dAcouMod_isdep, XC3dAcouMod_irdep, src, XC3dAcouMod_veldtinvdx2, izmax_inc, dat, wvmv );

  if (stat==1) 
     Errmsg("Error in XC3dWaveprop_op, program terminated!\n")   
  
  return 0;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////
int XC3dAcouMod_op_rndbnd(int tsign, int is, float *src, float *dat, float *wvmv)
{
  float *sdep,*rdep;
  int ix, iy;
  int nx, ny;
  int izmax_inc,stat;
  float dz;

  VerbMsg("Calculating sdep\n")
  sdep=&XC3dAcouMod_sdep[is*XC3dAcouMod_nx*XC3dAcouMod_ny];
  rdep=&XC3dAcouMod_rdep[is*XC3dAcouMod_nx*XC3dAcouMod_ny];
 
  nx=XC3dAcouMod_nx;
  ny=XC3dAcouMod_ny;
  dz=XC3dAcouMod_dz;
  // convert float depth (sdep, rdep) into integer index depth(isdep, irdep) 
 
  VerbMsg("Calculating isdep\n")
  for (iy=0;iy<ny;iy++)
     for (ix=0;ix<nx;ix++)
        XC3dAcouMod_isdep[ xyloc(ix,iy) ]=(sdep[ xyloc(ix,iy) ]-XC3dAcouMod_oz)/dz+XC3dAcouMod_nztape;

  VerbMsg("Calculating irdep\n")
  for (iy=0;iy<ny;iy++)
     for (ix=0;ix<nx;ix++)
        XC3dAcouMod_irdep[ xyloc(ix,iy) ]=(rdep[ xyloc(ix,iy) ]-XC3dAcouMod_oz)/dz+XC3dAcouMod_nztape;


  // wave propgation
  izmax_inc=XC3dAcouMod_izmax_inc;
  stat=XC3dWaveprop_op0nobnd( tsign, XC3dAcouMod_isdep, XC3dAcouMod_irdep, src, XC3dAcouMod_veldtinvdx2, izmax_inc, dat, wvmv );

  if (stat==1) 
     Errmsg("Error in XC3dWaveprop_op, program terminated!\n")   
  
  return 0;
}

int XC3dAcouMod_op_rndbnd1(int tsign, int is, float *src, float *dat, float *wvmv)
{
  float *sdep,*rdep;
  int ix, iy;
  int nx, ny;
  int izmax_inc,stat;
  float dz;

  VerbMsg("Calculating sdep\n")
  sdep=&XC3dAcouMod_sdep[is*XC3dAcouMod_nx*XC3dAcouMod_ny];
  rdep=&XC3dAcouMod_rdep[is*XC3dAcouMod_nx*XC3dAcouMod_ny];
 
  nx=XC3dAcouMod_nx;
  ny=XC3dAcouMod_ny;
  dz=XC3dAcouMod_dz;
  // convert float depth (sdep, rdep) into integer index depth(isdep, irdep) 
 
  VerbMsg("Calculating isdep\n")
  for (iy=0;iy<ny;iy++)
     for (ix=0;ix<nx;ix++)
        XC3dAcouMod_isdep[ xyloc(ix,iy) ]=(sdep[ xyloc(ix,iy) ]-XC3dAcouMod_oz)/dz+XC3dAcouMod_nztape;

  VerbMsg("Calculating irdep\n")
  for (iy=0;iy<ny;iy++)
     for (ix=0;ix<nx;ix++)
        XC3dAcouMod_irdep[ xyloc(ix,iy) ]=(rdep[ xyloc(ix,iy) ]-XC3dAcouMod_oz)/dz+XC3dAcouMod_nztape;


  // wave propgation
  izmax_inc=XC3dAcouMod_izmax_inc;
  stat=XC3dWaveprop_op2_rndbnd( XC3dAcouMod_isdep, XC3dAcouMod_irdep, src, XC3dAcouMod_veldtinvdx2, izmax_inc, dat, wvmv );

  if (stat==1) 
     Errmsg("Error in XC3dWaveprop_op, program terminated!\n")   
  
  return 0;
}
