#include "stdlib.h"
#include "math.h"
#include "string.h"
#include <seplib.h>
#include <sep3dc.h>


#define Errmsg3(str,str1,str2)  {fprintf(stderr,str,str1,str2);return 1;}
int main(int argc, char** argv)
{
    int ndim,i,n;
    float o,d;
    char temp_char[256];
    int   nxtape, nytape, nztape;
    float maxsigma, sigma, alpha_mean, alpha_sigma,alpha_maxV;
    int   nx, ny, nz;
    float ox, oy, oz;
    float dx, dy, dz;
    int   ix, iy, iz;
    int   ix1,iy1,iz1,i1;
    int   needwvmv,jtmv;
    int   stat, j_rndbnd;
    float sin, sinamp,pi,dv,wht;
    float *vel;
    int   *msk;
    float minV,maxV;
    pi=3.1415926;
    sep3d outfile; 
    initpar(argc,argv);
    // read t,x,y from source file
   
    if(0!=sep_get_number_data_axes("in",&ndim))
       seperr("unable to get dimension from src!\n");
    for(i=1;i<=ndim;i++)
    {
       if(0!=sep_get_data_axis_par("in",&i,&n,&o,&d,temp_char))
       seperr("unable to get parameters from src!\n");
       if(i==1)
       {
          nz=n;
          dz=d;
          oz=o;
       }
       else if(i==2)
       {
          nx=n;
          dx=d;
          ox=o;
       }
       else if(i==3)
       {
          ny=n;
          dy=d;
          oy=o;
       }
    }

    // readin other parameters
    if (0==getch("nxtape","i", &nxtape)) nxtape=0;
    if (0==getch("nytape","i", &nytape)) nytape=0;
    if (0==getch("nztape","i", &nztape)) nztape=0;

    if (0==getch("sigma","f", &sigma)) sigma=100;
    if (0==getch("maxsigma","f", &maxsigma)) maxsigma=10000;
    if (0==getch("alpha_mean","f", &alpha_mean)) alpha_mean=0.1;
    if (0==getch("alpha_sigma","f", &alpha_sigma)) alpha_sigma=0.1;
    if (0==getch("alpha_maxV","f", &alpha_maxV)) alpha_maxV=0.0;
    if (0==getch("sinamp","f", &sinamp)) sinamp=0.;

    if (0==getch("minV","f", &minV)) minV=1000;
    if (0==getch("maxV","f", &maxV)) maxV=2000;
    if (0==getch("j_rndbnd","i", &j_rndbnd)) j_rndbnd=1;

    if (minV>=maxV)
       Errmsg3("Error, minV %f greater than maxV %f \n",minV,maxV);

    // allocate memories

    vel=(float*)malloc(nx*ny*nz*sizeof(float));
    msk=(int*)malloc(nz*nx*ny*sizeof(int));

    srand(time(NULL));

    if ((nx*ny*nz*4)!=sreed("in",vel,nx*ny*nz*4))
    {
        fprintf(stderr,"Error read in vel file of size %d \n", nx*ny*nz*4);
        return 1;
    }

    // put 1 where velocity has been perturbed
    for (i=0;i<nx*ny*nz;i++)
       msk[i]=0;

    for (iy=nytape;iy<ny-nytape;iy++)
       for (ix=nxtape;ix<nx-nxtape;ix++)
          for (iz=nztape;iz<nz-nztape;iz++)
             msk[iz+ix*nz+iy*nz*nx]=1;

    // first pad all boundaries by extending values
    for (iy=nytape;iy<ny-nytape;iy++)
    {
       for (ix=0;ix<nxtape;ix++)
          for (iz=nztape;iz<nz-nztape;iz++)
          {
             vel[iz+ix*nz+iy*nz*nx]=vel[iz+nxtape*nz+iy*nz*nx];
             vel[iz+(nx-1-ix)*nz+iy*nz*nx]=vel[iz+(nx-1-nxtape)*nz+iy*nz*nx];
          }

       for (ix=0;ix<nx;ix++)
          for (iz=0;iz<nztape;iz++)
          {
             vel[iz+ix*nz+iy*nz*nx]=vel[nztape+ix*nz+iy*nz*nx];
             vel[(nz-1-iz)+ix*nz+iy*nz*nx]=vel[(nz-1-nztape)+ix*nz+iy*nz*nx];
          }
    }
 
    for (iy=0;iy<nytape;iy++)
       for (ix=0;ix<nx;ix++)
          for (iz=0;iz<nz;iz++)
          {
             vel[iz+ix*nz+iy*nz*nx]=vel[iz+ix*nz+nytape*nz*nx];
             vel[iz+ix*nz+(ny-1-iy)*nz*nx]=vel[iz+ix*nz+(ny-1-nytape)*nz*nx];
          }

       
    // then adjust the mean of boundary region seeds
    for (iy=0;iy<nytape;iy+=j_rndbnd)
       for (ix=0;ix<nx;ix+=j_rndbnd)
          for (iz=0;iz<nz;iz+=j_rndbnd)
          {
             vel[iz+ix*nz+iy*nz*nx]*=exp(-alpha_mean*(nytape-iy));
             vel[iz+ix*nz+(ny-1-iy)*nz*nx]*=exp(-alpha_mean*(nytape-iy));

          }

    for (iy=0;iy<ny;iy+=j_rndbnd)
       for (ix=0;ix<nxtape;ix+=j_rndbnd)
          for (iz=0;iz<nz;iz+=j_rndbnd)
          {
             vel[iz+ix*nz+iy*nz*nx]*=exp(-alpha_mean*(nxtape-ix));
             vel[iz+(nx-1-ix)*nz+iy*nz*nx]*=exp(-alpha_mean*(nxtape-ix));

          }

    for (iy=0;iy<ny;iy+=j_rndbnd)
       for (ix=0;ix<nx;ix+=j_rndbnd)
          for (iz=0;iz<nztape;iz+=j_rndbnd)
          {
             vel[iz+ix*nz+iy*nz*nx]*=exp(-alpha_mean*(nztape-iz));
             vel[(nz-1-iz)+ix*nz+iy*nz*nx]*=exp(-alpha_mean*(nztape-iz));

          }
        
    // then adjust the var of boundary region seeds
    for (iy=0;iy<nytape;iy+=j_rndbnd)
       for (ix=0;ix<nx;ix+=j_rndbnd)
          for (iz=0;iz<nz;iz+=j_rndbnd)
          {  
             i=iz+ix*nz+iy*nz*nx;
             vel[iz+ix*nz+iy*nz*nx]+=fminf(maxsigma,(sigma*exp(alpha_sigma*(nytape-iy))))*((2.0*rand()/RAND_MAX)-1.0);
             if (vel[i]>maxV/exp(alpha_maxV*(nytape-iy)))
                vel[i]=maxV/exp(alpha_maxV*(nytape-iy));
             msk[i]=1;
             i=iz+ix*nz+(ny-1-iy)*nz*nx;
             vel[iz+ix*nz+(ny-1-iy)*nz*nx]+=fminf(maxsigma, (sigma*exp(alpha_sigma*(nytape-iy))))*((2.0*rand()/RAND_MAX)-1.0);
             if (vel[i]>maxV/exp(alpha_maxV*(nytape-iy)))
                vel[i]=maxV/exp(alpha_maxV*(nytape-iy));
             msk[i]=1;

          }

    // then adjust the var of the whole boundary regions using nearest neighbor (statistically in all three directions)
    for (iy=0;iy<nytape;iy++)
       for (ix=0;ix<nx;ix++)
          for (iz=0;iz<nz;iz++)
          {  
             i=iz+ix*nz+iy*nz*nx;
             if (msk[i]==0)
             {
                iz1=((int)(iz/j_rndbnd))*j_rndbnd;
                ix1=((int)(ix/j_rndbnd))*j_rndbnd;
                iy1=((int)(iy/j_rndbnd))*j_rndbnd;
                if ((iz1+j_rndbnd)<nz)
                {
                   wht=1.*(iz-iz1)/j_rndbnd;
                   if (((1.0*rand())/RAND_MAX)<=wht)
                      iz1=iz1+j_rndbnd;
                } 
                if ((ix1+j_rndbnd)<nx)
                {
                   wht=1.*(ix-ix1)/j_rndbnd;
                   if (((1.0*rand())/RAND_MAX)<=wht)
                      ix1=ix1+j_rndbnd;
                } 
                if ((iy1+j_rndbnd)<nytape)
                {
                   wht=1.*(iy-iy1)/j_rndbnd;
                   if (((1.0*rand())/RAND_MAX)<=wht)
                      iy1=iy1+j_rndbnd;
                } 
                i1=iz1+ix1*nz+iy1*nz*nx;
                vel[i]=vel[i1];
             }

             i=iz+ix*nz+(ny-1-iy)*nz*nx;
             if (msk[i]==0)
             {
                iz1=((int)(iz/j_rndbnd))*j_rndbnd;
                ix1=((int)(ix/j_rndbnd))*j_rndbnd;
                iy1=((int)(iy/j_rndbnd))*j_rndbnd;
                if ((iz1+j_rndbnd)<nz)
                {
                   wht=1.*(iz-iz1)/j_rndbnd;
                   if (((1.0*rand())/RAND_MAX)<=wht)
                      iz1=iz1+j_rndbnd;
                } 
                if ((ix1+j_rndbnd)<nx)
                {
                   wht=1.*(ix-ix1)/j_rndbnd;
                   if (((1.0*rand())/RAND_MAX)<=wht)
                      ix1=ix1+j_rndbnd;
                } 
                if ((iy1+j_rndbnd)<nytape)
                {
                   wht=1.*(iy-iy1)/j_rndbnd;
                   if (((1.0*rand())/RAND_MAX)<=wht)
                      iy1=iy1+j_rndbnd;
                } 
                i1=iz1+ix1*nz+(ny-1-iy1)*nz*nx;
                vel[i]=vel[i1];
             }

          }

    // put 1 where velocity has been perturbed
    for (i=0;i<nx*ny*nz;i++)
       msk[i]=0;

    for (iy=nytape;iy<ny-nytape;iy++)
       for (ix=nxtape;ix<nx-nxtape;ix++)
          for (iz=nztape;iz<nz-nztape;iz++)
             msk[iz+ix*nz+iy*nz*nx]=1;

    // then adjust the var of boundary region seeds
    for (iy=0;iy<ny;iy+=j_rndbnd)
       for (ix=0;ix<nxtape;ix+=j_rndbnd)
          for (iz=0;iz<nz;iz+=j_rndbnd)
          {
             i=iz+ix*nz+iy*nz*nx;
             vel[iz+ix*nz+iy*nz*nx]+=fminf(maxsigma, (sigma*exp(alpha_sigma*(nxtape-ix))))*((2.0*rand()/RAND_MAX)-1.0);
             if (vel[i]>maxV/exp(alpha_maxV*(nxtape-ix)))
                vel[i]=maxV/exp(alpha_maxV*(nxtape-ix));
             msk[i]=1;
             i=iz+(nx-1-ix)*nz+iy*nz*nx;
             vel[i]+=fminf(maxsigma, (sigma*exp(alpha_sigma*(nxtape-ix))))*((2.0*rand()/RAND_MAX)-1.0);
             if (vel[i]>maxV/exp(alpha_maxV*(nxtape-ix)))
                vel[i]=maxV/exp(alpha_maxV*(nxtape-ix));
             msk[i]=1;
          }

    // then adjust the var of the whole boundary regions using nearest neighbor (statistically in all three directions)
    for (iy=0;iy<ny;iy++)
       for (ix=0;ix<nxtape;ix++)
          for (iz=0;iz<nz;iz++)
          {
             i=iz+ix*nz+iy*nz*nx;
             if (msk[i]==0)
             {
                iz1=((int)(iz/j_rndbnd))*j_rndbnd;
                ix1=((int)(ix/j_rndbnd))*j_rndbnd;
                iy1=((int)(iy/j_rndbnd))*j_rndbnd;
                if ((iz1+j_rndbnd)<nz)
                {
                   wht=1.*(iz-iz1)/j_rndbnd;
                   if (((1.0*rand())/RAND_MAX)<=wht)
                      iz1=iz1+j_rndbnd;
                } 
                if ((ix1+j_rndbnd)<nxtape)
                {
                   wht=1.*(ix-ix1)/j_rndbnd;
                   if (((1.0*rand())/RAND_MAX)<=wht)
                      ix1=ix1+j_rndbnd;
                } 
                if ((iy1+j_rndbnd)<ny)
                {
                   wht=1.*(iy-iy1)/j_rndbnd;
                   if (((1.0*rand())/RAND_MAX)<=wht)
                      iy1=iy1+j_rndbnd;
                } 
                i1=iz1+ix1*nz+iy1*nz*nx;
                vel[i]=vel[i1];
             }

             i=iz+(nx-1-ix)*nz+iy*nz*nx;
             if (msk[i]==0)
             {
                iz1=((int)(iz/j_rndbnd))*j_rndbnd;
                ix1=((int)(ix/j_rndbnd))*j_rndbnd;
                iy1=((int)(iy/j_rndbnd))*j_rndbnd;
                if ((iz1+j_rndbnd)<nz)
                {
                   wht=1.*(iz-iz1)/j_rndbnd;
                   if (((1.0*rand())/RAND_MAX)<=wht)
                      iz1=iz1+j_rndbnd;
                } 
                if ((ix1+j_rndbnd)<nxtape)
                {
                   wht=1.*(ix-ix1)/j_rndbnd;
                   if (((1.0*rand())/RAND_MAX)<=wht)
                      ix1=ix1+j_rndbnd;
                } 
                if ((iy1+j_rndbnd)<ny)
                {
                   wht=1.*(iy-iy1)/j_rndbnd;
                   if (((1.0*rand())/RAND_MAX)<=wht)
                      iy1=iy1+j_rndbnd;
                } 
                i1=iz1+(nx-1-ix1)*nz+iy1*nz*nx;
                vel[i]=vel[i1];
             }

          }

    // put 1 where velocity has been perturbed
    for (i=0;i<nx*ny*nz;i++)
       msk[i]=0;

    for (iy=nytape;iy<ny-nytape;iy++)
       for (ix=nxtape;ix<nx-nxtape;ix++)
          for (iz=nztape;iz<nz-nztape;iz++)
             msk[iz+ix*nz+iy*nz*nx]=1;

    // then adjust the var of boundary region seeds
    for (iy=0;iy<ny;iy+=j_rndbnd)
       for (ix=0;ix<nx;ix+=j_rndbnd)
          for (iz=0;iz<nztape;iz+=j_rndbnd)
          {
             i=iz+ix*nz+iy*nz*nx;
             vel[i]+=fminf(maxsigma, (sigma*exp(alpha_sigma*(nztape-iz))))*((2.0*rand()/RAND_MAX)-1.0);
             if (vel[i]>maxV/exp(alpha_maxV*(nztape-iz)))
                vel[i]=maxV/exp(alpha_maxV*(nztape-iz));
             msk[i]=1;
             i=(nz-1-iz)+ix*nz+iy*nz*nx;
             vel[i]+=fminf(maxsigma, (sigma*exp(alpha_sigma*(nztape-iz))))*((2.0*rand()/RAND_MAX)-1.0);
             if (vel[i]>maxV/exp(alpha_maxV*(nztape-iz)))
                vel[i]=maxV/exp(alpha_maxV*(nztape-iz));
             msk[i]=1;
          }

    // then adjust the var of the whole boundary regions using nearest neighbor (statistically in all three directions)
    
    for (iy=0;iy<ny;iy++)
       for (ix=0;ix<nx;ix++)
          for (iz=0;iz<nztape;iz++)
          {
             i=iz+ix*nz+iy*nz*nx;
             if (msk[i]==0)
             {
                iz1=((int)(iz/j_rndbnd))*j_rndbnd;
                ix1=((int)(ix/j_rndbnd))*j_rndbnd;
                iy1=((int)(iy/j_rndbnd))*j_rndbnd;
                if ((iz1+j_rndbnd)<nztape)
                {
                   wht=1.*(iz-iz1)/j_rndbnd;
                   if (((1.0*rand())/RAND_MAX)<=wht)
                      iz1=iz1+j_rndbnd;
                } 
                if ((ix1+j_rndbnd)<nx)
                {
                   wht=1.*(ix-ix1)/j_rndbnd;
                   if (((1.0*rand())/RAND_MAX)<=wht)
                      ix1=ix1+j_rndbnd;
                } 
                if ((iy1+j_rndbnd)<ny)
                {
                   wht=1.*(iy-iy1)/j_rndbnd;
                   if (((1.0*rand())/RAND_MAX)<=wht)
                      iy1=iy1+j_rndbnd;
                } 
                i1=iz1+ix1*nz+iy1*nz*nx;
                vel[i]=vel[i1];
             }

             i=(nz-1-iz)+ix*nz+iy*nz*nx;
             if (msk[i]==0)
             {
                iz1=((int)(iz/j_rndbnd))*j_rndbnd;
                ix1=((int)(ix/j_rndbnd))*j_rndbnd;
                iy1=((int)(iy/j_rndbnd))*j_rndbnd;
                if ((iz1+j_rndbnd)<nztape)
                {
                   wht=1.*(iz-iz1)/j_rndbnd;
                   if (((1.0*rand())/RAND_MAX)<=wht)
                      iz1=iz1+j_rndbnd;
                } 
                if ((ix1+j_rndbnd)<nx)
                {
                   wht=1.*(ix-ix1)/j_rndbnd;
                   if (((1.0*rand())/RAND_MAX)<=wht)
                      ix1=ix1+j_rndbnd;
                } 
                if ((iy1+j_rndbnd)<ny)
                {
                   wht=1.*(iy-iy1)/j_rndbnd;
                   if (((1.0*rand())/RAND_MAX)<=wht)
                      iy1=iy1+j_rndbnd;
                } 
                i1=(nz-1-iz1)+ix1*nz+iy1*nz*nx;
                vel[i]=vel[i1];
             }

          }

    for (i=0;i<nx*ny*nz;i++)
    {
       if (vel[i]<minV)
          vel[i]=minV;
    } 

    ndim=3;
    if(0!=init_sep3d_par(&outfile,"OUTPUT","FLOAT","REGULAR",ndim,0))
       seperr("unable to write outfile!\n");
    outfile.o[0]=oz;
    outfile.d[0]=dz;
    outfile.n[0]=nz;
    outfile.o[1]=ox;
    outfile.d[1]=dx;
    outfile.n[1]=nx;
    outfile.o[2]=oy;
    outfile.d[2]=dy;
    outfile.n[2]=ny;
    sep3d_set_sep3d(&outfile);
    if(0!=sep3dc_write_description("vel",&outfile))
       seperr("unable to output parameters for vel\n");

    if( (nz*nx*ny*4)!=srite("vel",vel,4*nz*nx*ny) )
       seperr("trouble writing out vel\n");

    return 0;
}
