#define Errmsg2(str,str1)  {fprintf(stderr,str,str1);return 1;}
#define Errmsg3(str,str1,str2)  {fprintf(stderr,str,str1,str2);return 1;}

#include "stdlib.h"
#include "math.h"
#include "string.h"
#include <seplib.h>
#include <sep3dc.h>
#include "XC3dSolver.c"


int main(int argc, char** argv)
{
    int ndim,i,n;
    float o,d;
    char temp_char[256];
    int   nt, nx, ny, nz, ns;
    float ot, ox, oy, oz;
    float dt, dx, dy, dz;
    int   it, ix, iy, iz, is;
    int   nxtape,nytape,nztape;
    int   needwvmv,jtmv,needprecond,niter;
    float vmin, vmax, mmin, mmax;
    int   sz,stat;
    float *vel, *src, *dobs, *dwht, *sdep, *rdep, *wvmv, *resd;
    sep3d outfile; 
    time_t time1,time2;

    initpar(argc,argv);

    // readin other parameters
    if (0==getch("nxtape","i", &nxtape)) nxtape=50;
    if (0==getch("nytape","i", &nytape)) nytape=50;
    if (0==getch("nztape","i", &nztape)) nztape=50;
    if (0==getch("jtmv","i", &jtmv)) jtmv=10;
    if (0==getch("vmin","f", &vmin)) vmin=1200;
    if (0==getch("vmax","f", &vmax)) vmax=2000;
    if (0==getch("needprecond","i", &needprecond)) needprecond=0;
    if (0==getch("niter","i", &niter)) niter=1;

    // read t,x,y from data file
    if(0!=sep_get_number_data_axes("in",&ndim))
       seperr("unable to get dimension from dat!\n");
    
    for(i=1;i<=ndim;i++)
    {
       if(0!=sep_get_data_axis_par("in",&i,&n,&o,&d,temp_char))
       seperr("unable to get parameters from dat!\n");
       if(i==1)
       {
          nt=n;
          dt=d;
          ot=o;
       }
       else if(i==2)
       {
          nx=n+2*nxtape;
          dx=d;
          ox=o-d*nxtape;
       }
       else if(i==3)
       {
          ny=n+2*nytape;
          dy=d;
          oy=o-d*nytape;
       }
       if (ndim==3) 
          ns=1; 
       else if(i==4)
          ns=n;
    }

    // read z info from vel
    if(0!=sep_get_number_data_axes("src",&ndim))
       seperr("unable to get dimension from src file!\n");
    for(i=1;i<=ndim;i++)
    {
       if(0!=sep_get_data_axis_par("src",&i,&n,&o,&d,temp_char))
       seperr("unable to get parameters from src file!\n");
       if(i==1)
       {
          if (nt!=n)
          Errmsg3("nt in data %d different from nt %d in src, program terminated!\n",nt,n)
          if (dt!=d)
          Errmsg3("dt in data %f different from dt %f in src, program terminated!\n",dt,d)
          if (ot!=o)
          Errmsg3("ot in data %f different from ot %f in src, program terminated!\n",ot,o)
       }
       else if(i==2)
       {
          if (nx!=n+2*nxtape)
          Errmsg3("nx in data %d different from nx %d in src, program terminated!\n",nx,n)
          if (dx!=d)
          Errmsg3("dx in data %f different from dx %f in src, program terminated!\n",dx,d)
          if (ox!=o-d*nxtape)
          Errmsg3("ox in data %f different from ox %f in src, program terminated!\n",ox,o)
       }
       else if(i==3)
       {
          if (ny!=n+2*nytape)
          Errmsg3("ny in data %d different from ny %d in src, program terminated!\n",ny,n)
          if (dy!=d)
          Errmsg3("dy in data %f different from dy %f in src, program terminated!\n",dy,d)
          if (oy!=o-d*nytape)
          Errmsg3("oy in data %f different from oy %f in src, program terminated!\n",oy,o)
       }
       if (ndim==3) 
          ns=1; 
       else if(i==4)
          if (ns!=n)
             Errmsg3("ns in dat %d different from ns %d in src, program terminated!\n",ns,n)
    }

    // read z info from vel
    if(0!=sep_get_number_data_axes("vel",&ndim))
       seperr("unable to get dimension from vel file!\n");
    for(i=1;i<=ndim;i++)
    {
       if(0!=sep_get_data_axis_par("vel",&i,&n,&o,&d,temp_char))
       seperr("unable to get parameters from vel file!\n");
       if(i==1)
       {
          nz=n;
          dz=d;
          oz=o;
       }
       else if(i==2)
       {
          if (nx!=n)
          Errmsg3("nx in vel %d different from nx %d in src, program terminated!\n",nx,n)
          if (dx!=d)
          Errmsg3("dx in vel %f different from dx %f in src, program terminated!\n",dx,d)
          if (ox!=o)
          Errmsg3("ox in vel %f different from ox %f in src, program terminated!\n",ox,o)
       }
       else if(i==3)
       {
          if (ny!=n)
          Errmsg3("ny in vel %d different from ny %d in src, program terminated!\n",ny,n)
          if (dy!=d)
          Errmsg3("dy in vel %f different from dy %f in src, program terminated!\n",dy,d)
          if (oy!=o)
          Errmsg3("oy in vel %f different from oy %f in src, program terminated!\n",oy,o)
       }
    }

    if (nx<=2*nxtape)
       Errmsg3("Error, nx %d less or equal than twice of padded x reigon %d \n",nx,nxtape);
    if (ny<=2*nytape)
       Errmsg3("Error, ny %d less or equal than twice of padded y reigon %d \n",ny,nytape);
    if (nz<=2*nztape)
       Errmsg3("Error, nz %d less or equal than twice of padded z reigon %d \n",nz,nztape);

    // allocate memories

    vel=(float*)malloc(nx*ny*nz*sizeof(float));
    sz=(nx-2*nxtape)*(ny-2*nytape)*nt*ns;
    VerbMsg1("----------sz of src is %d-------------\n",sz)
    src=(float*)malloc(sz*sizeof(float));
    dobs=(float*)malloc(sz*sizeof(float));
    dwht=(float*)malloc(sz*sizeof(float));
    sz=(nx-2*nxtape)*(ny-2*nytape)*ns;
    sdep=(float*)malloc(sz*sizeof(float));
    rdep=(float*)malloc(sz*sizeof(float));
    resd=(float*)malloc(niter*sizeof(float));

    sz=(nx-2*nxtape)*(ny-2*nytape)*nt*ns;
    // read in various data files into arrays
    if ((sz*4)!=sreed("in",dobs,sz*4))
       Errmsg2("Error read in dat file of size %d \n",sz*4);

    if ((sz*4)!=sreed("src",src,sz*4))
       Errmsg2("Error read in src file of size %d \n",sz*4);

    VerbMsg("start reading vel\n")
    if ((nx*ny*nz*4)!=sreed("vel",vel,nx*ny*nz*4))
       Errmsg2("Error read in vel file of size %d \n", nx*ny*nz*4);

    VerbMsg("start reading sdep\n")
    FILE *fp;
    fp=fopen("sdep","r");
    sz=(nx-2*nxtape)*(ny-2*nytape)*ns;
    if (fp!=NULL)
    {
       fclose(fp);
       if ((sz*4)!=sreed("sdep",sdep,sz*4))
          seperr("Error read in sdep file\n");
    }
    else
    {
       for (i=0;i<sz;i++)
          sdep[i]=2*dz;
    }

    VerbMsg("start reading rdep\n")
    fp=fopen("rdep","r");
    sz=(nx-2*nxtape)*(ny-2*nytape)*ns;
    if (fp)
    {
       fclose(fp);
       if ((sz*4)!=sreed("rdep",rdep,sz*4))
          seperr("Error read in rdep file\n");
    }
    else
    {
       for (i=0;i<sz;i++)
          rdep[i]=2*dz;
    }

    VerbMsg("start reading dwht\n")
    sz=(nx-2*nxtape)*(ny-2*nytape)*nt*ns;
    fp=fopen("dwht","r");
    if (fp!=NULL)
    {
       fclose(fp);
       if ((sz*4)!=sreed("dwht",dwht,sz*4))
          seperr("Error read in dwht file\n");
    }
    else
    {
       for (i=0;i<sz;i++)
          dwht[i]=1;
    }

    // Initialize wave propagation 
    VerbMsg("start initialize XC3dSolver_init\n")
    mmin=(dx/(vmax*dt))*(dx/(vmax*dt));
    mmax=(dx/(vmin*dt))*(dx/(vmin*dt));
    stat=XC3dSolver_init( nt,nx,ny,nz,ns,       \
                          ot,ox,oy,oz,          \
                          dt,dx,dy,dz,          \
                          nxtape,nytape,nztape, \
                          sdep, rdep, jtmv, mmin, mmax, needprecond );

    if (stat==1)
       seperr("Error in XC3dSolver_init,program terminated!\n");

    sz=nx*ny*nz;
    for (i=0;i<sz;i++)
       vel[i]=(dx/(dt*vel[i]))*(dx/(dt*vel[i]));  
    VerbMsg("start XC3dSolver_op\n")
    stat=XC3dSolver_op(vel, dobs, dwht, src, niter, resd);

    for (i=0;i<sz;i++)
       vel[i]=dx/(dt*(sqrtf(vel[i])));  

    //output resd
    ndim=1;
    if(0!=init_sep3d_par(&outfile,"OUTPUT","FLOAT","REGULAR",ndim,0))
       seperr("unable to write outfile!\n");
    outfile.o[0]=0;
    outfile.d[0]=1;
    outfile.n[0]=niter;
    sep3d_set_sep3d(&outfile);
    if(0!=sep3dc_write_description("resd",&outfile))
       seperr("unable to output parameters for resd\n");

    if( (niter*4)!=srite("resd",resd,4*niter) )
       seperr("trouble writing out resd\n");

    //output updated velocity
    ndim=3;
    if(0!=init_sep3d_par(&outfile,"OUTPUT","FLOAT","REGULAR",ndim,0))
       seperr("unable to write outfile!\n");
    outfile.o[0]=oz;
    outfile.d[0]=dz;
    outfile.n[0]=nz;
    outfile.o[1]=ox;
    outfile.d[1]=dx;
    outfile.n[1]=nx;
    outfile.o[2]=oy;
    outfile.d[2]=dy;
    outfile.n[2]=ny;
    outfile.n[3]=ns;
    sep3d_set_sep3d(&outfile);
    if(0!=sep3dc_write_description("vout",&outfile))
       seperr("unable to output parameters for vout\n");

    if( (nz*nx*ny*4)!=srite("vout",vel,4*nz*nx*ny) )
       seperr("trouble writing out vout\n");

    return 0;
}
