#include "stdio.h"
#include "stdlib.h"

#include "XC3dWaveprop.h"
#include "3dstencil.h"
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
int XC3dWaveprop_Init(int ntin,  int nxin,  int nyin,  int nzin, \
                      int nxtapein, int nytapein, int nztapein,  \
                      int needwvmvin, int jtmvin, int freesurf )
{
   int sizet;
   XC3dWaveprop_nt=ntin;
   XC3dWaveprop_nx=nxin;
   XC3dWaveprop_ny=nyin;
   XC3dWaveprop_nz=nzin;
   XC3dWaveprop_nxtape=nxtapein;
   XC3dWaveprop_nytape=nytapein;
   XC3dWaveprop_nztape=nztapein;
   XC3dWaveprop_needdat=1;
   XC3dWaveprop_needwvmv=needwvmvin;
   XC3dWaveprop_jtmv=jtmvin;
   XC3dWaveprop_freesurf=freesurf;
 
   //VerbMsg1("needwvmv in Waveprop is %d\n",needwvmvin) 
   // determine if padded reigon is reasonable
   if (nxin<=2*nxtapein)
      WpErrmsg("Error, nx less or equal than twice of padded x reigon\n")
   if (nyin<=2*nytapein)
      WpErrmsg("Error, ny less or equal than twice of padded y reigon\n")
   if (nzin<=2*nztapein)
      WpErrmsg("Error, nz less or equal than twice of padded z reigon\n")


   sizet=nxin*nyin*nzin;
   XC3dWaveprop_wavet1=(float*)malloc(sizet*sizeof(float));
   XC3dWaveprop_wavet2=(float*)malloc(sizet*sizeof(float));
   XC3dWaveprop_wavet3=(float*)malloc(sizet*sizeof(float));
   XC3dWaveprop_tmpwv=(float*)malloc(sizet*sizeof(float));

  VerbMsg1("!!!!!!!!!!!!!!!!!!! in Waveprop2 nx=%d \n",nxin) 
  VerbMsg1("!!!!!!!!!!!!!!!!!!! in Waveprop2 ny=%d \n",nyin) 
  VerbMsg1("!!!!!!!!!!!!!!!!!!! in Waveprop2 nz=%d \n",nzin) 
  VerbMsg1("!!!!!!!!!!!!!!!!!!! in Waveprop2 nxtape=%d \n",nxtapein) 
  VerbMsg1("!!!!!!!!!!!!!!!!!!! in Waveprop2 nytape=%d \n",nytapein) 
  VerbMsg1("!!!!!!!!!!!!!!!!!!! in Waveprop2 nztape=%d \n",nztapein) 
  VerbMsg1("!!!!!!!!!!!!!!!!!!! in Waveprop2 nt=%d \n",ntin) 
  VerbMsg1("!!!!!!!!!!!!!!!!!!! in Waveprop2 jt=%d \n",jtmvin) 
   //if (XC3dWaveprop_needwvmv>0)
   //   XC3dWaveprop_wvmv=(float*)malloc(sizet*sizeof(float)*(ntin/jtmvin));
   

   return 0;
}

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
//izmax_inc : incremental z range of wavefield to be calculated as t increases
int XC3dWaveprop_op(int tsign, int *isdep, int *irdep, float *src, float *veldtinvdx2, int izmax_inc, float *data, float *wvmv )
{
  int ix,iy,iz,it,jt;
  int nx,ny,nz,nt;
  int stat;
  float *wvtmp;
  int izmin,izmax;
  time_t time1,time2;
  nt=XC3dWaveprop_nt;
  jt=XC3dWaveprop_jtmv;
  nx=XC3dWaveprop_nx;
  ny=XC3dWaveprop_ny;
  nz=XC3dWaveprop_nz;
  // zero out all the wavefield
  for (iy=0;iy<ny;iy++)
     for (ix=0;ix<nx;ix++)
        for (iz=0;iz<nz;iz++)
           XC3dWaveprop_wavet1[ Wpzxyloc(iz,ix,iy) ]=0.;
  
  for (iy=0;iy<ny;iy++)
     for (ix=0;ix<nx;ix++)
        for (iz=0;iz<nz;iz++)
           XC3dWaveprop_wavet2[ Wpzxyloc(iz,ix,iy) ]=0.;

  for (iy=0;iy<ny;iy++)
     for (ix=0;ix<nx;ix++)
        for (iz=0;iz<nz;iz++)
           XC3dWaveprop_wavet3[ Wpzxyloc(iz,ix,iy) ]=0.;

  VerbMsg("start propagation in Waveprop_op with \n") 
  // start propagation
  time1=time(NULL);
  for (it=0;it<nt;it++)
  {   
     // inject source  
     VerbMsg("start inject source Waveprop_op\n") 
     stat=XC3dWaveprop_inject(it,isdep,src);
     if (stat==1) 
        WpErrmsg("Error in XC3dWaveprop_inject, program teminated!\n");

     // record data  
     if (XC3dWaveprop_needdat>0)
     {
        VerbMsg("start record data in Waveprop_op\n") 
        if (tsign>0)  
           stat=XC3dWaveprop_rcd(it,irdep,data);
        else  
           stat=XC3dWaveprop_rcd(nt-it-1,irdep,data);
        if (stat==1) 
           WpErrmsg("Error in XC3dWaveprop_rcd, program teminated!\n");
     }

     // record wave movie if necessary  
     if ( (XC3dWaveprop_needwvmv>0) && (it%jt==0) ) 
     { 
        VerbMsg("start record movie in Waveprop_op\n") 
        if (tsign>0)  
           stat=XC3dWaveprop_rcdmv(it,wvmv);
        else
           stat=XC3dWaveprop_rcdmv(nt-it-1,wvmv);

        if (stat==1) 
           WpErrmsg("Error in XC3dWaveprop_rcdmv, program teminated!\n");
     }
 
     // determine z range for wave propagation    
     izmin=nz;
     izmax=0; 

     for(iy=1;iy<ny;iy++)
        for(ix=1;ix<nx;ix++)
        {
           iz=isdep[Wpxyloc(ix,iy)];
           if (iz>izmax)
              izmax=iz;
           if (iz<izmin)
              izmin=iz;
        }

     if (XC3dWaveprop_freesurf==0)
        izmin=izmin-it*izmax_inc;
     else
        izmin=XC3dWaveprop_nztape; 
     izmax=izmax+it*izmax_inc;
   
     // 1 step propagation
     VerbMsg("start 1 step propagation in Waveprop_op \n") 
     //stat=XC3dWaveprop_1step4t10spacewithbnd1opt(veldtinvdx2, izmin, izmax);
     stat=XC3dWaveprop_1step4t10spacewithbnd1opt(veldtinvdx2, izmin, izmax);
     if (stat==1) 
        WpErrmsg("Error in XC3dWaveprop_1step2t2space, program teminated!\n");

     // swap wavefield slices
     wvtmp=XC3dWaveprop_wavet1;
     XC3dWaveprop_wavet1=XC3dWaveprop_wavet2;
     XC3dWaveprop_wavet2=wvtmp;
     //XC3dWaveprop_wavet2=XC3dWaveprop_wavet3;
     //XC3dWaveprop_wavet3=wvtmp;
     
     if ( (it%(nt/10))==0 ) 
     {
        time2=time(NULL);
        //VerbMsg1("Finished %d percent of one shot modeling ",(it*100/nt))
        //VerbMsg1("spent %ld seconds\n",(time2-time1))
        time1=time(NULL);
     }
  }

  return 0;
}

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
//izmax_inc : incremental z range of wavefield to be calculated as t increases
int XC3dWaveprop_op0(int tsign, int *isdep, int *irdep, float *src, float *veldtinvdx2, int izmax_inc, float *data, float *wvmv )
{
  int ix,iy,iz,it,jt;
  int nx,ny,nz,nt;
  int stat;
  float *wvtmp;
  float *tmparray;
  float *xval, *yval, *zval;
  int izmin,izmax;
  int ixmin,ixmax;
  int iymin,iymax;
  int flag;
  time_t time1,time2,totaltime;
  nt=XC3dWaveprop_nt;
  jt=XC3dWaveprop_jtmv;
  nx=XC3dWaveprop_nx;
  ny=XC3dWaveprop_ny;
  nz=XC3dWaveprop_nz;

  tmparray=(float*)malloc(125*sizeof(float));
  xval=(float*)malloc(11*sizeof(float));
  yval=(float*)malloc(11*sizeof(float));
  zval=(float*)malloc(11*sizeof(float));
  // zero out all the wavefield
  for (iy=0;iy<ny;iy++)
     for (ix=0;ix<nx;ix++)
        for (iz=0;iz<nz;iz++)
           XC3dWaveprop_wavet1[ Wpzxyloc(iz,ix,iy) ]=0.;
  
  for (iy=0;iy<ny;iy++)
     for (ix=0;ix<nx;ix++)
        for (iz=0;iz<nz;iz++)
           XC3dWaveprop_wavet2[ Wpzxyloc(iz,ix,iy) ]=0.;

  for (iy=0;iy<ny;iy++)
     for (ix=0;ix<nx;ix++)
        for (iz=0;iz<nz;iz++)
           XC3dWaveprop_wavet3[ Wpzxyloc(iz,ix,iy) ]=0.;
  // determine iymin
  flag=0;
  for (iy=0;iy<ny;iy++)
  {
     if (flag==0)
     {
        for (ix=0;ix<nx;ix++)
        {
           if (flag==0) 
           {
              for (it=0;it<nt;it++)
                 if ( src[Wptxyloc(it,ix,iy)] != 0.)
                 {
                    iymin=iy;
                    flag=1;
                    break;
                 }
           }
        }
     }
  }
  // determine iymax
  flag=0;
  for (iy=ny-1;iy>-1;iy--)
  {
     if (flag==0)
     {
        for (ix=0;ix<nx;ix++)
        {
           if (flag==0) 
           {
              for (it=0;it<nt;it++)
                 if ( src[Wptxyloc(it,ix,iy)] != 0.)
                 {
                    iymax=iy;
                    flag=1;
                    break;
                 }
           }
        }
     }
  }
  // determine ixmin
  flag=0;
  for (ix=0;ix<nx;ix++)
  {
     if (flag==0)
     {
        for (iy=0;iy<ny;iy++)
        {
           if (flag==0) 
           {
              for (it=0;it<nt;it++)
                 if ( src[Wptxyloc(it,ix,iy)] != 0.)
                 {
                    ixmin=ix;
                    flag=1;
                    break;
                 }
           }
        }
     }
  }
  //determine ixmax
  flag=0;
  for (ix=nx-1;ix>-1;ix--)
  {
     if (flag==0)
     {
        for (iy=0;iy<ny;iy++)
        {
           if (flag==0) 
           {
              for (it=0;it<nt;it++)
                 if ( src[Wptxyloc(it,ix,iy)] != 0.)
                 {
                    ixmax=ix;
                    flag=1;
                    break;
                 }
           }
        }
     }
  }
  VerbMsg("start propagation in Waveprop_op with \n") 
  // start propagation
  totaltime=0.;
  for (it=0;it<nt;it++)
  {   
     // inject source  
     VerbMsg("start inject source Waveprop_op\n") 
     if (tsign>0)  
        stat=XC3dWaveprop_inject(it,isdep,src);
     else  
        stat=XC3dWaveprop_inject(nt-it-1,isdep,src);
     if (stat==1) 
        WpErrmsg("Error in XC3dWaveprop_inject, program teminated!\n");

     // record data  
     if (XC3dWaveprop_needdat>0)
     {
        VerbMsg("start record data in Waveprop_op\n") 
        if (tsign>0)  
           stat=XC3dWaveprop_rcd(it,irdep,data);
        else  
           stat=XC3dWaveprop_rcd(nt-it-1,irdep,data);
        if (stat==1) 
           WpErrmsg("Error in XC3dWaveprop_rcd, program teminated!\n");
     }

     // record wave movie if necessary  
     if ( (XC3dWaveprop_needwvmv>0) && (it%jt==0) ) 
     { 
        VerbMsg("start record movie in Waveprop_op\n") 
        if (tsign>0)  
           stat=XC3dWaveprop_rcdmv(it,wvmv);
        else
           stat=XC3dWaveprop_rcdmv(nt-it-1,wvmv);

        if (stat==1) 
           WpErrmsg("Error in XC3dWaveprop_rcdmv, program teminated!\n");
     }
 
     // determine z range for wave propagation    
     izmin=nz;
     izmax=0; 

     for(iy=1;iy<ny;iy++)
        for(ix=1;ix<nx;ix++)
        {
           iz=isdep[Wpxyloc(ix,iy)];
           if (iz>izmax)
              izmax=iz;
           if (iz<izmin)
              izmin=iz;
        }
     if (XC3dWaveprop_freesurf==0)
        izmin=izmin-it*izmax_inc;
     else
        izmin=XC3dWaveprop_nztape; 

     izmax=izmax+it*izmax_inc;
     ixmin=ixmin-izmax_inc;
     ixmax=ixmax+izmax_inc;
     iymin=iymin-izmax_inc;
     iymax=iymax+izmax_inc;
   
     // 1 step propagation
     VerbMsg("start 1 step propagation in Waveprop_op \n") 
     //stat=XC3dWaveprop_1step4t10spacewithbnd1opt(veldtinvdx2, izmin, izmax);
     //time1=time(NULL);
     //stat=XC3dWaveprop_1step4t10spacewithbnd1opt2(veldtinvdx2, izmin, izmax, ixmin, ixmax, iymin, iymax);
     stat=XC3dWaveprop_1step4t10spacewithbnd1opt2(veldtinvdx2, izmin, izmax, ixmin, ixmax, iymin, iymax, \
                                                  XC3dWaveprop_nxtape, XC3dWaveprop_nytape, XC3dWaveprop_nztape, \
                                                  XC3dWaveprop_nx, XC3dWaveprop_ny, XC3dWaveprop_nz, \
                                                  XC3dWaveprop_wavet1, XC3dWaveprop_wavet2, XC3dWaveprop_wavet3, \
                                                  XC3dWaveprop_tmpwv);
     //time2=time(NULL); 
    //totaltime+=time2-time1;
     //stat=XC3dWaveprop_1step4t10spacewithbnd1opt1(veldtinvdx2, izmin, izmax, tmparray, xval, yval, zval );
     if (stat==1) 
        WpErrmsg("Error in XC3dWaveprop_1step2t2space, program teminated!\n");

     // swap wavefield slices
     wvtmp=XC3dWaveprop_wavet1;
     XC3dWaveprop_wavet1=XC3dWaveprop_wavet2;
     XC3dWaveprop_wavet2=wvtmp;
     //XC3dWaveprop_wavet2=XC3dWaveprop_wavet3;
     //XC3dWaveprop_wavet3=wvtmp;
     
  }
  //fprintf(stderr,"used time for propagation is %ld seconds\n", totaltime);
  free(tmparray);
  free(xval);
  free(yval);
  free(zval);
  return 0;
}

//izmax_inc : incremental z range of wavefield to be calculated as t increases
int XC3dWaveprop_op0nobnd(int tsign, int *isdep, int *irdep, float *src, float *veldtinvdx2, int izmax_inc, float *data, float *wvmv )
{
  int ix,iy,iz,it,jt;
  int nx,ny,nz,nt;
  int stat;
  float *wvtmp;
  float *tmparray;
  float *xval, *yval, *zval;
  int izmin,izmax;
  int ixmin,ixmax;
  int iymin,iymax;
  int flag;
  time_t time1,time2,totaltime;
  nt=XC3dWaveprop_nt;
  jt=XC3dWaveprop_jtmv;
  nx=XC3dWaveprop_nx;
  ny=XC3dWaveprop_ny;
  nz=XC3dWaveprop_nz;

  tmparray=(float*)malloc(125*sizeof(float));
  xval=(float*)malloc(11*sizeof(float));
  yval=(float*)malloc(11*sizeof(float));
  zval=(float*)malloc(11*sizeof(float));
  // zero out all the wavefield
  for (iy=0;iy<ny;iy++)
     for (ix=0;ix<nx;ix++)
        for (iz=0;iz<nz;iz++)
           XC3dWaveprop_wavet1[ Wpzxyloc(iz,ix,iy) ]=0.;
  
  for (iy=0;iy<ny;iy++)
     for (ix=0;ix<nx;ix++)
        for (iz=0;iz<nz;iz++)
           XC3dWaveprop_wavet2[ Wpzxyloc(iz,ix,iy) ]=0.;

  for (iy=0;iy<ny;iy++)
     for (ix=0;ix<nx;ix++)
        for (iz=0;iz<nz;iz++)
           XC3dWaveprop_wavet3[ Wpzxyloc(iz,ix,iy) ]=0.;
  // determine iymin
  flag=0;
  for (iy=0;iy<ny;iy++)
  {
     if (flag==0)
     {
        for (ix=0;ix<nx;ix++)
        {
           if (flag==0) 
           {
              for (it=0;it<nt;it++)
                 if ( src[Wptxyloc(it,ix,iy)] != 0.)
                 {
                    iymin=iy;
                    flag=1;
                    break;
                 }
           }
        }
     }
  }
  // determine iymax
  flag=0;
  for (iy=ny-1;iy>-1;iy--)
  {
     if (flag==0)
     {
        for (ix=0;ix<nx;ix++)
        {
           if (flag==0) 
           {
              for (it=0;it<nt;it++)
                 if ( src[Wptxyloc(it,ix,iy)] != 0.)
                 {
                    iymax=iy;
                    flag=1;
                    break;
                 }
           }
        }
     }
  }
  // determine ixmin
  flag=0;
  for (ix=0;ix<nx;ix++)
  {
     if (flag==0)
     {
        for (iy=0;iy<ny;iy++)
        {
           if (flag==0) 
           {
              for (it=0;it<nt;it++)
                 if ( src[Wptxyloc(it,ix,iy)] != 0.)
                 {
                    ixmin=ix;
                    flag=1;
                    break;
                 }
           }
        }
     }
  }
  //determine ixmax
  flag=0;
  for (ix=nx-1;ix>-1;ix--)
  {
     if (flag==0)
     {
        for (iy=0;iy<ny;iy++)
        {
           if (flag==0) 
           {
              for (it=0;it<nt;it++)
                 if ( src[Wptxyloc(it,ix,iy)] != 0.)
                 {
                    ixmax=ix;
                    flag=1;
                    break;
                 }
           }
        }
     }
  }
  VerbMsg("start propagation in Waveprop_op with \n") 
  // start propagation
  totaltime=0.;
  for (it=0;it<nt;it++)
  {  
   //fprintf(stderr,"start it=%d\n",it);
     // inject source  
     VerbMsg("start inject source Waveprop_op\n") 
     stat=XC3dWaveprop_inject(it,isdep,src);
     if (stat==1) 
        WpErrmsg("Error in XC3dWaveprop_inject, program teminated!\n");

     // record data  
     if (XC3dWaveprop_needdat>0)
     {
        VerbMsg("start record data in Waveprop_op\n") 
        if (tsign>0)  
           stat=XC3dWaveprop_rcd(it,irdep,data);
        else  
           stat=XC3dWaveprop_rcd(nt-it-1,irdep,data);
        if (stat==1) 
           WpErrmsg("Error in XC3dWaveprop_rcd, program teminated!\n");
     }

   //fprintf(stderr,"start rcdmv it=%d, tsign=%d\n",it,tsign);
     // record wave movie if necessary  
     if ( (XC3dWaveprop_needwvmv>0) && (it%jt==0) ) 
     { 
        VerbMsg("start record movie in Waveprop_op\n") 
        if (tsign>0)  
           stat=XC3dWaveprop_rcdmv(it,wvmv);
        else
           stat=XC3dWaveprop_rcdmv(nt-it-1,wvmv);

        if (stat==1) 
           WpErrmsg("Error in XC3dWaveprop_rcdmv, program teminated!\n");
     }
 
     //fprintf(stderr,"done rcdmv it=%d\n",it);
     // determine z range for wave propagation    
     izmin=nz;
     izmax=0; 

     for(iy=1;iy<ny;iy++)
        for(ix=1;ix<nx;ix++)
        {
           iz=isdep[Wpxyloc(ix,iy)];
           if (iz>izmax)
              izmax=iz;
           if (iz<izmin)
              izmin=iz;
        }

     if (XC3dWaveprop_freesurf==0)
        izmin=izmin-it*izmax_inc;
     else
        izmin=XC3dWaveprop_nztape; 

     izmax=izmax+it*izmax_inc;
     ixmin=ixmin-izmax_inc;
     ixmax=ixmax+izmax_inc;
     iymin=iymin-izmax_inc;
     iymax=iymax+izmax_inc;
   
     // 1 step propagation
     VerbMsg("start 1 step propagation in Waveprop_op \n") 
     //stat=XC3dWaveprop_1step4t10spacewithbnd1opt(veldtinvdx2, izmin, izmax);
     //time1=time(NULL);
     //stat=XC3dWaveprop_1step4t10spacewithbnd1opt2(veldtinvdx2, izmin, izmax, ixmin, ixmax, iymin, iymax);
     stat=XC3dWaveprop_1step4t10spaceopt2(veldtinvdx2, izmin, izmax, ixmin, ixmax, iymin, iymax, \
                                          XC3dWaveprop_nxtape, XC3dWaveprop_nytape, XC3dWaveprop_nztape, \
                                          XC3dWaveprop_nx, XC3dWaveprop_ny, XC3dWaveprop_nz, \
                                          XC3dWaveprop_wavet1, XC3dWaveprop_wavet2, XC3dWaveprop_wavet3, \
                                          XC3dWaveprop_tmpwv);
     //time2=time(NULL); 
    //totaltime+=time2-time1;
     //stat=XC3dWaveprop_1step4t10spacewithbnd1opt1(veldtinvdx2, izmin, izmax, tmparray, xval, yval, zval );
     if (stat==1) 
        WpErrmsg("Error in XC3dWaveprop_1step2t2space, program teminated!\n");

     // swap wavefield slices
     wvtmp=XC3dWaveprop_wavet1;
     XC3dWaveprop_wavet1=XC3dWaveprop_wavet2;
     XC3dWaveprop_wavet2=wvtmp;
     //XC3dWaveprop_wavet2=XC3dWaveprop_wavet3;
     //XC3dWaveprop_wavet3=wvtmp;
     
  }
  //fprintf(stderr,"used time for propagation is %ld seconds\n", totaltime);
  free(tmparray);
  free(xval);
  free(yval);
  free(zval);
  return 0;
}

//izmax_inc : incremental z range of wavefield to be calculated as t increases
int XC3dWaveprop_op2(int tsign, int *isdep, int *irdep, float *src, float *dxinvvelinvdt2, int izmax_inc, float *data, float *wvmv )
{
  int ix,iy,iz,it,jt;
  int nx,ny,nz,nt;
  int nxtape,nytape;
  int stat;
  float *wvtmp;
  int izmin,izmax;
  time_t time1,time2;
  nt=XC3dWaveprop_nt;
  jt=XC3dWaveprop_jtmv;
  nx=XC3dWaveprop_nx;
  ny=XC3dWaveprop_ny;
  nxtape=XC3dWaveprop_nxtape;
  nytape=XC3dWaveprop_nytape;
  nz=XC3dWaveprop_nz;
  // zero out all the wavefield
  for (iy=0;iy<ny;iy++)
     for (ix=0;ix<nx;ix++)
        for (iz=0;iz<nz;iz++)
           XC3dWaveprop_wavet1[ Wpzxyloc(iz,ix,iy) ]=0.;
  
  for (iy=0;iy<ny;iy++)
     for (ix=0;ix<nx;ix++)
        for (iz=0;iz<nz;iz++)
           XC3dWaveprop_wavet2[ Wpzxyloc(iz,ix,iy) ]=0.;

  for (iy=0;iy<ny;iy++)
     for (ix=0;ix<nx;ix++)
        for (iz=0;iz<nz;iz++)
           XC3dWaveprop_wavet3[ Wpzxyloc(iz,ix,iy) ]=0.;

  WpVerbMsg("start propagation in Waveprop_op2 with \n") 
  // start propagation
  time1=time(NULL);
  for (it=0;it<nt;it++)
  {   
     // inject source  
     WpVerbMsg("start inject source Waveprop_op2\n") 
     if (tsign>0)  
        stat=XC3dWaveprop_injectnopad(it,isdep,src);
     else  
        stat=XC3dWaveprop_injectnopad(nt-it-1,irdep,src);
     if (stat==1) 
        WpErrmsg("Error in XC3dWaveprop_injectnopad, program teminated!\n");

     // record data  
     if (XC3dWaveprop_needdat>0)
     {
        WpVerbMsg("start record data in Waveprop_op2\n") 
        if (tsign>0)  
           stat=XC3dWaveprop_rcdnopad(it,irdep,data);
        else  
           stat=XC3dWaveprop_rcdnopad(nt-it-1,irdep,data);
        if (stat==1) 
           WpErrmsg("Error in XC3dWaveprop_rcdnopad, program teminated!\n");
     }

     // record wave movie if necessary  
     if ( (XC3dWaveprop_needwvmv>0) && (it%jt==0) ) 
     { 
        WpVerbMsg("start record movie in Waveprop_op2\n") 
        if (tsign>0)  
           stat=XC3dWaveprop_rcdmvnopad(it,wvmv);
        else
           stat=XC3dWaveprop_rcdmvnopad(((int)((nt-1)/jt))*jt-it,wvmv);

        if (stat==1) 
           WpErrmsg("Error in XC3dWaveprop_rcdmvnopad, program teminated!\n");
     }
 
     // determine z range for wave propagation    
     izmin=nz;
     izmax=0; 

     for(iy=1;iy<ny-2*nytape;iy++)
        for(ix=1;ix<nx-2*nxtape;ix++)
        {
           iz=isdep[Wpxylocnopad(ix,iy)];
           if (iz>izmax)
              izmax=iz;
           if (iz<izmin)
              izmin=iz;
        }

     if (XC3dWaveprop_freesurf==0)
        izmin=izmin-it*izmax_inc;
     else
        izmin=XC3dWaveprop_nztape; 

     izmax=izmax+it*izmax_inc;
   
     // 1 step propagation
     WpVerbMsg("start 1 step propagation in Waveprop_op2 \n") 
     stat=XC3dWaveprop_1step4t10spacewithbnd1optinvvelin(dxinvvelinvdt2, izmin, izmax);
     if (stat==1) 
        WpErrmsg("Error in XC3dWaveprop_1step4t10spacewithbnd1optinline, program teminated!\n");

     // swap wavefield slices
     WpVerbMsg("start pointer shift in Waveprop_op2 \n") 
     wvtmp=XC3dWaveprop_wavet1;
     XC3dWaveprop_wavet1=XC3dWaveprop_wavet2;
     XC3dWaveprop_wavet2=wvtmp;
     //XC3dWaveprop_wavet2=XC3dWaveprop_wavet3;
     //XC3dWaveprop_wavet3=wvtmp;
     
     if ( (it%(nt/10))==0 ) 
     {
        time2=time(NULL);
        //VerbMsg1("Finished %d percent of one shot modeling ",(it*100/nt))
        //VerbMsg1("spent %ld seconds\n",(time2-time1))
        time1=time(NULL);
     }
  }

  return 0;
}


///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
int XC3dWaveprop_setrcddat(int needdat)
{
   XC3dWaveprop_needdat=needdat;
   return 0;
}

int XC3dWaveprop_setwvmv(int needwvmvin)
{
   XC3dWaveprop_needwvmv=needwvmvin;
   return 0;
}
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
int XC3dWaveprop_inject(int it, int *isdep, float *src)
{
   int ix,iy,iz;
   int nx,ny;
   int nxtape,nytape;
   nx=XC3dWaveprop_nx;
   ny=XC3dWaveprop_ny;
   nxtape=XC3dWaveprop_nxtape;
   nytape=XC3dWaveprop_nytape;
   for (iy=0;iy<ny;iy++)
      for (ix=0;ix<nx;ix++)
      {
         iz=isdep[ Wpxyloc(ix,iy) ];
         XC3dWaveprop_wavet2[ Wpzxyloc(iz,ix,iy) ] +=src[ Wptxyloc(it,ix,iy) ];
      }
   return 0;
}

int XC3dWaveprop_injectnopad(int it, int *isdep, float *src)
{
   int ix,iy,iz;
   int nx,ny;
   int nxtape,nytape;
   nx=XC3dWaveprop_nx;
   ny=XC3dWaveprop_ny;
   nxtape=XC3dWaveprop_nxtape;
   nytape=XC3dWaveprop_nytape;
   for (iy=0;iy<ny-2*nytape;iy++)
      for (ix=0;ix<nx-2*nxtape;ix++)
      {
         iz=isdep[ Wpxylocnopad(ix,iy) ];
         XC3dWaveprop_wavet2[ Wpzxyloc(iz,ix+nxtape,iy+nytape) ] +=src[ Wptxylocnopad(it,ix,iy) ];
      }
   return 0;
}

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
int XC3dWaveprop_rcd(int it, int *irdep, float *data)
{
   int ix,iy,iz;
   int nx,ny;
   int nxtape,nytape;
   nx=XC3dWaveprop_nx;
   ny=XC3dWaveprop_ny;
   nxtape=XC3dWaveprop_nxtape;
   nytape=XC3dWaveprop_nytape;
   for (iy=0;iy<ny;iy++)
      for (ix=0;ix<nx;ix++)
      {
         iz=irdep[ Wpxyloc(ix,iy) ];
         data[ Wptxyloc(it,ix,iy) ] = XC3dWaveprop_wavet2[ Wpzxyloc(iz,ix,iy) ];
      }
   return 0;
}

int XC3dWaveprop_rcdnopad(int it, int *irdep, float *data)
{
   int ix,iy,iz;
   int nx,ny;
   int nxtape,nytape;
   nx=XC3dWaveprop_nx;
   ny=XC3dWaveprop_ny;
   nxtape=XC3dWaveprop_nxtape;
   nytape=XC3dWaveprop_nytape;
   for (iy=0;iy<ny-2*nytape;iy++)
      for (ix=0;ix<nx-2*nxtape;ix++)
      {
         iz=irdep[ Wpxylocnopad(ix,iy) ];
         data[ Wptxylocnopad(it,ix,iy) ] = XC3dWaveprop_wavet2[ Wpzxyloc(iz,ix+nxtape,iy+nytape) ];
      }
   return 0;
}
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
int XC3dWaveprop_rcdmv(int it, float *wvmv)
{
   int ix,iy,iz;
   int jt;
   int nx,ny,nz;
   nx=XC3dWaveprop_nx;
   ny=XC3dWaveprop_ny;
   nz=XC3dWaveprop_nz;
   jt=XC3dWaveprop_jtmv;
   //fprintf(stderr,"nx=%d, ny=%d, nz=%d, jt=%d, it=%d\n",nx,ny,nz,jt,it);
   for (iy=0;iy<ny;iy++)
      for (ix=0;ix<nx;ix++)
         for (iz=0;iz<nz;iz++)
         {
            /*
            fprintf(stderr,"ix=%d, iy=%d, iz=%d,jt=%d,it=%d, loc=%d, input=%f\n",ix,iy,iz,jt,it,Wpzxyloc(iz,ix,iy), XC3dWaveprop_wavet1[ Wpzxyloc(iz,ix,iy) ]);
            fprintf(stderr,"outindex=%d\n",(Wpzxyloc(iz,ix,iy) + nx*ny*nz*(it/jt)));
            fprintf(stderr,"output=%f\n",wvmv[ Wpzxyloc(iz,ix,iy) + nx*ny*nz*(it/jt) ]);*/
            wvmv[ Wpzxyloc(iz,ix,iy) + nx*ny*nz*(it/jt) ] = XC3dWaveprop_wavet2[ Wpzxyloc(iz,ix,iy) ];
         }
            //XC3dWaveprop_wvmv[ Wpzxyloc(iz,ix,iy) + nx*ny*nz*(it/jt) ] = XC3dWaveprop_wavet1[ Wpzxyloc(iz,ix,iy) ];
   return 0;
}

int XC3dWaveprop_rcdmvnopad(int it, float *wvmv)
{
   int ix,iy,iz;
   int jt;
   int nx,ny,nz;
   int nxtape,nytape,nztape;
   nx=XC3dWaveprop_nx;
   ny=XC3dWaveprop_ny;
   nz=XC3dWaveprop_nz;
   nxtape=XC3dWaveprop_nxtape;
   nytape=XC3dWaveprop_nytape;
   nztape=XC3dWaveprop_nztape;
   jt=XC3dWaveprop_jtmv;
   for (iy=0;iy<ny-2*nytape;iy++)
      for (ix=0;ix<nx-2*nxtape;ix++)
         for (iz=0;iz<nz-2*nztape;iz++)
            wvmv[ Wpzxylocnopad(iz,ix,iy) + (nx-2*nxtape)*(ny-2*nytape)*(nz-2*nztape)*(it/jt) ] = XC3dWaveprop_wavet2[ Wpzxyloc(iz+nztape,ix+nxtape,iy+nytape) ];
            //XC3dWaveprop_wvmv[ Wpzxyloc(iz,ix,iy) + nx*ny*nz*(it/jt) ] = XC3dWaveprop_wavet1[ Wpzxyloc(iz,ix,iy) ];
   return 0;
}

//izmax_inc : incremental z range of wavefield to be calculated as t increases
int XC3dWaveprop_op2_rndbnd(int *isdep, int *irdep, float *src, float *veldtinvdx2, int izmax_inc, float *data, float *wvmv  )
{
  int ix,iy,iz,it,jt;
  int nx,ny,nz,nt;
  int nxtape,nytape;
  int stat;
  float *wvtmp;
  int izmin,izmax;
  int ixmin,ixmax;
  int iymin,iymax;
  int flag;
  time_t time1,time2;
  nt=XC3dWaveprop_nt;
  jt=XC3dWaveprop_jtmv;
  nx=XC3dWaveprop_nx;
  ny=XC3dWaveprop_ny;
  nxtape=XC3dWaveprop_nxtape;
  nytape=XC3dWaveprop_nytape;
  nz=XC3dWaveprop_nz;
  // zero out all the wavefield
  VerbMsg("start XC3dWaveprop_op2_rndbnd_imag\n")
  WpVerbMsg("start zero XC3dWaveprop_wavet1\n")
  for (iy=0;iy<ny;iy++)
     for (ix=0;ix<nx;ix++)
        for (iz=0;iz<nz;iz++)
        {
           XC3dWaveprop_wavet1[ Wpzxyloc(iz,ix,iy) ]=0.;
           XC3dWaveprop_wavet2[ Wpzxyloc(iz,ix,iy) ]=0.;
           XC3dWaveprop_wavet3[ Wpzxyloc(iz,ix,iy) ]=0.;
           XC3dWaveprop_tmpwv[ Wpzxyloc(iz,ix,iy) ]=0.;
        }
                                                       
  WpVerbMsg("start propagation in Waveprop_op2_rndbnd_imag with \n")
  // start propagation
  time1=time(NULL);
  // determine iymin
  flag=0;
  for (iy=0;iy<ny;iy++)
  {
     if (flag==0)
     {
        for (ix=0;ix<nx;ix++)
        {
           if (flag==0)
           {
              for (it=0;it<nt;it++)
                 if ( src[Wptxyloc(it,ix,iy)] != 0.)
                 {
                    iymin=iy;
                    flag=1;
                    break;
                 }
           }
        }
     }
  }
  // determine iymax
  flag=0;
  for (iy=ny-1;iy>-1;iy--)
  {
     if (flag==0)
     {
        for (ix=0;ix<nx;ix++)
        {
           if (flag==0)
           {
              for (it=0;it<nt;it++)
                 if ( src[Wptxyloc(it,ix,iy)] != 0.)
                 {
                    iymax=iy;
                    flag=1;
                    break;
                 }
           }
        }
     }
                                                                 }
 // determine ixmin
  flag=0;
  for (ix=0;ix<nx;ix++)
  {
     if (flag==0)
     {
        for (iy=0;iy<ny;iy++)
        {
           if (flag==0)
           {
              for (it=0;it<nt;it++)
                 if ( src[Wptxyloc(it,ix,iy)] != 0.)
                 {
                    ixmin=ix;
                    flag=1;
                    break;
                 }
           }
        }
     }
  }
  //determine ixmax
  flag=0;
  for (ix=nx-1;ix>-1;ix--)
  {
     if (flag==0)
     {
        for (iy=0;iy<ny;iy++)
        {
           if (flag==0)
           {
              for (it=0;it<nt;it++)
                 if ( src[Wptxyloc(it,ix,iy)] != 0.)
                 {
                    ixmax=ix;
                    flag=1;
                    break;
                 }
           }
        }
     }
  }


  // determine z range for wave propagation    
  izmin=nz;
  izmax=0;

  for(iy=0;iy<ny;iy++)
     for(ix=0;ix<nx;ix++)
     {
        iz=isdep[Wpxyloc(ix,iy)];
        if (iz>izmax)
           izmax=iz;
       if (iz<izmin)
           izmin=iz;
     }

  for (it=0;it<nt;it++)
  {
     VerbMsg1("in Waveprop_op2, start time step %d\n",it)
     // inject source  
     WpVerbMsg("start inject source Waveprop_op2\n")
     stat=XC3dWaveprop_inject(it,isdep,src);
     //fprintf(stderr,"in forward, it=%d\n",it);
     if (stat==1)
        WpErrmsg("Error in XC3dWaveprop_inject, program teminated!\n");

     if (XC3dWaveprop_freesurf==0)
        izmin=izmin-izmax_inc;
     else
        izmin=XC3dWaveprop_nztape; 

     izmax=izmax+izmax_inc;
     ixmin=ixmin-izmax_inc;
     ixmax=ixmax+izmax_inc;
     iymin=iymin-izmax_inc;
     iymax=iymax+izmax_inc;


     // 1 step propagation
     stat=XC3dWaveprop_1step4t10spaceopt2(veldtinvdx2, izmin, izmax, ixmin, ixmax, iymin, iymax, \
                                          XC3dWaveprop_nxtape, XC3dWaveprop_nytape, XC3dWaveprop_nztape, \
                                          XC3dWaveprop_nx, XC3dWaveprop_ny, XC3dWaveprop_nz, \
                                          XC3dWaveprop_wavet1, XC3dWaveprop_wavet2, XC3dWaveprop_wavet3, \
                                          XC3dWaveprop_tmpwv);
     if (stat==1)
        WpErrmsg("Error in XC3dWaveprop_1step4t10spacewithbnd1optinline, program teminated!\n");

     stat=XC3dWaveprop_rcd(it,irdep,data);
     if ( (XC3dWaveprop_needwvmv>0) && (it%jt==0) ) 
     { 
        stat=XC3dWaveprop_rcdmv(it,wvmv);
        if (stat==1) 
           WpErrmsg("Error in XC3dWaveprop_rcdmv, program teminated!\n");
     }
     // swap wavefield slices
     WpVerbMsg("start pointer shift in Waveprop_op2_rndbnd \n")
     wvtmp=XC3dWaveprop_wavet1;
     XC3dWaveprop_wavet1=XC3dWaveprop_wavet2;
     XC3dWaveprop_wavet2=wvtmp;
     //wvtmp=XC3dWaveprop_wavet1;
     //XC3dWaveprop_wavet1=XC3dWaveprop_wavet2;
     //XC3dWaveprop_wavet2=XC3dWaveprop_wavet3;
     //XC3dWaveprop_wavet3=wvtmp;
  }

  wvtmp=XC3dWaveprop_wavet1;
  XC3dWaveprop_wavet1=XC3dWaveprop_wavet2;
  XC3dWaveprop_wavet2=wvtmp;

  wvtmp=XC3dWaveprop_wavet3;
  XC3dWaveprop_wavet3=XC3dWaveprop_wavet1;
  XC3dWaveprop_wavet2=XC3dWaveprop_wavet2;
  XC3dWaveprop_wavet1=wvtmp;
  //wvtmp=XC3dWaveprop_wavet3;
  //XC3dWaveprop_wavet3=XC3dWaveprop_wavet2;
  //XC3dWaveprop_wavet2=XC3dWaveprop_wavet1;
  //XC3dWaveprop_wavet1=wvtmp;

  for (it=nt-1;it>-1;it--)
  {
     VerbMsg1("in Waveprop_op2_rndbnd_imag, start time step %d\n",it)

     if (XC3dWaveprop_freesurf==0)
        izmin=izmin-izmax_inc;
     else
        izmin=XC3dWaveprop_nztape; 

     izmax=izmax+izmax_inc;
     ixmin=ixmin-izmax_inc;
     ixmax=ixmax+izmax_inc;
     iymin=iymin-izmax_inc;
     iymax=iymax+izmax_inc;


     stat=XC3dWaveprop_1step4t10spaceopt2(veldtinvdx2, izmin, izmax, ixmin, ixmax, iymin, iymax, \
                                          XC3dWaveprop_nxtape, XC3dWaveprop_nytape, XC3dWaveprop_nztape, \
                                          XC3dWaveprop_nx, XC3dWaveprop_ny, XC3dWaveprop_nz, \
                                          XC3dWaveprop_wavet3, XC3dWaveprop_wavet2, XC3dWaveprop_wavet1, \
                                          XC3dWaveprop_tmpwv);
     if (stat==1)
        WpErrmsg("Error in XC3dWaveprop_1step4t10spaceoptinline, program teminated!\n");
     /*
     if ( (XC3dWaveprop_needwvmv>0) && (it%jt==0) ) 
     { 
        stat=XC3dWaveprop_rcdmv(it,wvmv);
        if (stat==1) 
           WpErrmsg("Error in XC3dWaveprop_rcdmv, program teminated!\n");
     }*/

     // swap wavefield slices
     WpVerbMsg("start pointer shift in Waveprop_op2_rndbnd \n")
     wvtmp=XC3dWaveprop_wavet3;
     XC3dWaveprop_wavet3=XC3dWaveprop_wavet2;
     XC3dWaveprop_wavet2=wvtmp;
     //XC3dWaveprop_wavet2=XC3dWaveprop_wavet1;
     //XC3dWaveprop_wavet1=wvtmp;

  }

  return 0;
}


